/*
 * This code is based on the x10.shar program that circulates the net
 * and is often references in the comp.home.automation newsgroup.
 * I have simplified it significantly by removing the language and all
 * the features that are unnecessary (such as sunset, timers, etc) for
 * strict on/off computer control.
 * %W%	%G%	- Doug Hughes - Auburn University College of Engineering
 *
 * This assumes the existance of /dev/x10 which should be a symbollic
 * link to the serial device where you CP290 is attached. You may
 * use the -d flag to change it, or just change the Device global below.
 *
 * You may use this program for any purpose provided you leave this
 * banner at the top - doug@eng.auburn.edu
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <setjmp.h>
#include <termio.h>
#include <sys/fcntl.h>

#define  u(X)            (1 << (16-X))

static char *Device="/dev/x10";
#ifndef Debug
int Debug = 0;
#endif
static jmp_buf env;

usage()
{
	puts("usage: x10 [-d device] -z <zone> -c <code#> -a <on|off>");
	exit(1);
}
	
/* initialize tty for 600 baud  and cbreak style char at a time processing */
set_tty(fd)
{
    int fd0;
    struct termios tp;
 
    if (tcgetattr(fd, &tp) < 0)
        return(-1);
    tp.c_iflag &= ~(IXON|IXOFF|ISTRIP);
    tp.c_oflag &= ~(OPOST);
    tp.c_cflag = B600|CS8|CLOCAL;
    tp.c_lflag &= ~(ISIG|ICANON);
    tp.c_cc[VMIN] = 1;
    tp.c_cc[VTIME] = 0;
 
    return (tcsetattr(fd, TCSAFLUSH, &tp));
}

x10_flush(int fd)
{
	tcflush(fd, TCIOFLUSH);
}

/*
 *  Mapping arrays and subroutines, to/from X10 codes
 */
unsigned char   zone2code[16] =
{
    0x60, 0xe0, 0x20, 0xa0, 0x10, 0x90, 0x50, 0xd0,
    0x70, 0xf0, 0x30, 0xb0, 0x00, 0x80, 0x40, 0xc0
};

/* map numeric codes to letter names */
char	code2zone[16] =
{
    'm', 'e', 'c', 'k',  'o', 'g', 'a', 'i',
    'n', 'f', 'd', 'l',  'p', 'h', 'b', 'j'
};


x10_read(int fd, unsigned char *buf, int len)
{
    register int    rc;
 
    while (len)
    {
        rc = read(fd, buf, Debug >= 2 ? 1 : len);
        if (Debug >= 2)
            printf("rc = %d %x\n", rc, buf[0]);
        if (rc < 0) return (rc);
        len -= rc;
        buf += rc;
    }
	return(len);
}

int x10_ack(int fd)
{
    unsigned char   buf[7+1];
    register int    len;
 
    len = x10_read(fd, &buf[1], 7);
    if (Debug >= 2)
        printf("len = %d, ACK status=%d\n", len, buf[7]);
}
 
void handlejmp(int d)
{
	longjmp(env, 1);
}

/* upload a command from cp290 */
x10_cmd_up(int fd)
{
    unsigned char   buf[12+1];
    register int    len;
 
    len = x10_read(fd, &buf[1], 12);
    if (Debug >= 2)
        printf("len = %d, status=%d\n", len, buf[7]);
	return(buf[8]);
}
 

/* Set the base zone code */
int x10_set_basezone(int fd, int zone)
{
	unsigned char b[18+1];
	
	memset((char *) &b[1], 0xff, 18);
	b[17] = 0;
	b[18] = zone2code[zone-'a'];
	write(fd, &b[1], 18);
	if (Debug)
		printf("code sent is %x\n", b[18]);
}

/* send command to cp290 - see cp290 book for details */
int x10_send_command(int fd, int zone, int codes, int action) 
{
	unsigned char b[22+1];
	unsigned char sum = 0;
	int b8;
	
	memset((char *) b, 0xff, 16+1);

	b[17] = 1;
	sum += b[18] = action;
	sum += b[19] = zone2code[zone];
	sum += b[20] = codes;
	sum += b[21] = codes >> 8;
	b[22] = sum;
	write(fd, &b[1], 22);
	if (Debug)
		printf("code sent is %x\n", b[18]);

	alarm(2);
	signal(SIGALRM, handlejmp);
	if (setjmp(env) == 0) {
		x10_ack(fd);
		b8 = x10_cmd_up(fd);
	} else {
#ifndef _TCL
		printf("incomplete read from interface.\n");
#endif
		return(1);
	}
	alarm(0);
	if (code2zone[b8>>4] != zone+'a' || (b8 & 0x0f) != action) {
#ifndef _TCL
		printf("code recvd is %c,%x, expected %c,%x\n", code2zone[b8>>4], 
				b8&0xf, zone+'a', b[18] &0xf);
#endif
		return(1);
	}

	return(0);
}


#ifndef _TCL
main(int argc, char *argv[])
{
	int c;
	extern int optind, opterr, optopt;
	extern char *optarg;
	char *zp;			/* zone pointer */
	char *cp;			/* device code pointer */
	int zone;
	int action = 2;
	int tcode;
	int code;
	int fd;

	while ((c = getopt(argc, argv, "d:z:c:a:?h")) != EOF) {
		switch(c) {
			case 'd':	
				Device = strdup(optarg);
				break;
			case '?':
			case 'h':
			default:
				usage();
				break;
		}
	}
	argc -= optind;
	argv += optind;


	if (argc < 3) 
		usage();

	if (( fd = open(Device, O_RDWR)) < 0) {
		perror("Opening x10 device");
		exit(1);
	}

	if (set_tty(fd) < 0) {
        perror("Couldn't set term attributes");
        exit(1);
    }

	zp = argv[0];
	action =  (strcmp(argv[2], "off") == 0 ? 3 : 2);
	

	while (*zp != NULL) {
		zone = (int) *zp & 0x7f;
		if (zone < 'a' || zone > 'p') {
			puts("zone must be housecode from 'a' to 'p'");
			exit(1);
		}
		printf ("doing zone %c\n", zone);
		cp = argv[1];
		code = 0;
		while (isdigit(*cp)) {
			tcode = atoi(cp);
			if (!(tcode > 0 && tcode <= 16)) {
				puts("Code must be integer value from 1 to 16");
				exit(1);
			}
			printf("code is %d\n", tcode);
			while (isdigit(*cp++));
			if (*cp == ',') cp++;
			code |= u(tcode);
		}
		if (code == 0) {
			puts("Code must be integer or comma separated integers [1-16]");
			exit(1);
		}

		if (*++zp == ',') zp++; /* skip commas, get next zone */
		printf("sending %c %d %d\n", zone, code, action);
		if (x10_send_command(fd, zone-'a', code, action))
			printf("failure for %c %d %d\n", zone, code, action);
	}
		


	exit(0);
}

#endif


